# -*- coding:utf-8 -*-
import kafka
from datetime import datetime
import subscribe_pb2 as subscribe
import to_sql

PARTITION_SEQ_KEY = "ps"
SHARD_ID = "ShardId"


def tdsql_block_consume(brokers, topic, group, user, password, trans2sql):
    # create a kafka consumer
    consumer = kafka.KafkaConsumer(topic,
                                   auto_offset_reset='earliest',
                                   enable_auto_commit=False,
                                   group_id=group,
                                   sasl_plain_username=user,
                                   sasl_plain_password=password,
                                   sasl_mechanism='SCRAM-SHA-512',
                                   security_protocol='SASL_PLAINTEXT',
                                   bootstrap_servers=brokers)

    # get partition of topic
    partitions = consumer.partitions_for_topic(topic)
    partition_consumers = {}

    for p in partitions:
        partition_consumers[p] = PartitionMsgConsumer()

    for message in consumer:
        partition_consumers[message.partition].consume(consumer, message, trans2sql)


def get_partition_seq(message):
    for m in message.headers:
        if m[0] == PARTITION_SEQ_KEY:
            return int(m[1].decode())


def get_shard_id(message):
    for m in message.headers:
        if m[0] == SHARD_ID:
            return str(m[1].decode())


def timestamp2str(timestamp):
    date_time = datetime.fromtimestamp(timestamp)
    return date_time.strftime("%Y-%m-%d %H:%M:%S")


def on_entry(partition, offset, partition_seq, entry, trans2sql):
    if trans2sql:
        header = entry.header
        print(
            "-->[partition: {}, offset: {}, partitionSeq: {}] [{}:{}], happenedAt: {}".format(
                partition, offset, partition_seq, header.fileName, header.position, timestamp2str(header.timestamp))
        )

        return to_sql.trans2sql(entry)
    else:
        print(
            "-->[partition: {}, offset: {}, partitionSeq: {}]\n{}".format(
                partition, offset, partition_seq, entry)
        )


def unmarshal_envelope(bindata):
    envelope = subscribe.Envelope()
    envelope.ParseFromString(bindata)
    return envelope


def unmarshal_entries(bindata):
    entries = subscribe.Entries()
    entries.ParseFromString(bindata)
    return entries


class PartitionMsgConsumer(object):
    def __init__(self):
        self.cur_complete_msg = subscribe.Envelope()
        self.shard_last_seq_map = {}
        self.shard_msg_map = {}
        self.last_partition_seq = 0

    def consume(self, consumer, message, trans2sql):

        # Each message has a sequence id in the partition, this id is continuous monotonic increasing,
        # these messages are duplicated if the id rotate;
        # There is unexpected message lost in kafka if the id is discontinuous,
        # have to reset the subscribe and client when it happened.
        # When the task is restarted, the sequence id will start from 1
        ps = get_partition_seq(message)

        # Each message has a shard id,
        # this id can be used to distinguish that the message belongs to which shard
        shard_id = get_shard_id(message)
        shard_last_seq = self.shard_last_seq_map.get(shard_id, -1)
        if shard_last_seq == -1 or ps == 1:
            self.shard_last_seq_map[shard_id] = 1
        elif ps <= shard_last_seq:
            print("duplicated msg:{}".format(message))
        elif ps != shard_last_seq + 1 and shard_last_seq != 0:
            msg = "the partition seq discontinuous, last: {}, current: {}".format(self.last_partition_seq, ps)
            raise Exception(msg)

        self.shard_last_seq_map[shard_id] = ps
        envelope = unmarshal_envelope(message.value)

        if envelope.version != 1:
            print("unsupported version: {}, message: {}".format(envelope.version, envelope))
            raise Exception("unsupported version")

        # The size of one event maybe exceed the size limitation of kafka message,
        # then the event will be split into more than one message,
        # each message has an index, the value starts from zero to total-1,
        # have to rebuild the byte array before decode the object by protobuf.
        if envelope.total == 1:
            self.cur_complete_msg = envelope
        else:
            shard_msg = self.shard_msg_map.get(shard_id)
            if shard_msg is None:
                self.shard_msg_map[shard_id] = envelope
            else:
                shard_msg.data += envelope.data
                self.shard_msg_map[shard_id] = shard_msg

        if envelope.index < envelope.total - 1:
            return

        if envelope.total == 1:
            entries = unmarshal_entries(self.cur_complete_msg.data)
        else:
            entries = unmarshal_entries(self.shard_msg_map[shard_id].data)
            # Pay attention to memory recycling for shard_msg_map
            del self.shard_msg_map[shard_id]

        for entry in entries.items:
            errmsg = on_entry(message.partition, message.offset, ps, entry, trans2sql)
            if errmsg:
                raise Exception(errmsg)

        # commit at checkpoint message
        if entries.items and entries.items[0].header.messageType == subscribe.MessageType.CHECKPOINT:
            offsets = {
                kafka.TopicPartition(message.topic, message.partition): kafka.OffsetAndMetadata(message.offset+1, None),
            }
            consumer.commit(offsets)
